/*
 *
 * CPORTGetInt.C
 *
 * The calling syntax is:
 *
 *		[Y,err]=cportgetint(H,nbyte,S,C)
 *
 */

#include <mex.h>
#include <windows.h>
#include <limits.h>

/* Input Arguments */

#define	PORT_IN	 prhs[0]
#define NBYTE_IN prhs[1]
#define S_IN     prhs[2]
#define C_IN     prhs[3]

/* Output Arguments */

#define Y_OUT    plhs[0]
#define	STATUS   plhs[1]

long num2longtsr(
                 const double x,
                 const int upper,
                 const int lower
                )
{
/*
 *
 * num2longtsr
 *
 *  Saturates a value if it goes above an upper bound or below a lower bound.
 *  Rounds to nearest long integer value.
 *
 */
    if (x < lower)
        return lower;
    else if (x > upper)
        return upper;
    else 
        return (long)(x<0 ? x - 0.5 : x + 0.5);

}


void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		        )
{
    HANDLE	*hp;
    double	*Stat, *Pr, *Pi;
    char    *ch;
    BOOL    ok, complex;
    long    word;
    DWORD   nBytes, nRead, nCharToRead, nToRead, nDims, *dims, i, word_ind, byte_ind, ch_ind;

    /* Check for proper number of arguments */

    if ((nrhs < 1) | (nrhs > 4))
        mexErrMsgTxt( "CPORTGETINT requires 1 to 4 input arguments." );
    if ((nlhs < 1) | (nlhs > 2)) 
        mexErrMsgTxt( "CPORTGETINT requires 1 or 2 output arguments." );
  
    /* Check the type of PORT_IN and assign pointer */
    if (!mxIsDouble(PORT_IN))
        mexErrMsgTxt("CPORTGETINT first argument should be a file handle.");
    hp = (HANDLE *)mxGetData( PORT_IN );

    /* if nbyte specified */
    if (nrhs > 1)
    {
        if (!mxIsDouble( NBYTE_IN ))
            mexErrMsgTxt( "CPORTGETINT second argument should be a number." );
        Pr = mxGetPr( NBYTE_IN );
        nBytes = num2longtsr( *Pr, sizeof( long ), 1  );
    }
    else
        nBytes = 2;

    /* if S specified */
    if (nrhs > 2)
    {
        if (!mxIsDouble( S_IN ))
            mexErrMsgTxt( "CPORTGETINT third argument should be a number." );
        nDims   = num2longtsr( mxGetNumberOfElements( S_IN ), LONG_MAX, 1 );

        Pr = mxGetPr( S_IN );
        dims = mxMalloc( (nDims +1)*sizeof( double ) );
        dims[0] = num2longtsr( Pr[0], LONG_MAX, 0 );
        nToRead = dims[0];
        for (i = 1; i < nDims; i++)
        {
            dims[i] = num2longtsr( Pr[i], LONG_MAX, 0 );
            nToRead *= dims[i];
        }
        if (nDims == 1)
        {
            nDims = 2;                                          /* generate column vector if singleton size is specified */
            dims[1] = 1;
        }
    }
    else
    {
        nDims   = 2;
        dims = mxMalloc( nDims*sizeof(double) );
        dims[0] = 1;
        dims[1] = 1;
        nToRead = 1;
    }

    /* if C specified */
    if (nrhs > 3)
    {
        if (!mxIsDouble( C_IN ))
            mexErrMsgTxt( "CPORTGETINT fourth argument should be a number." );
        Pr   = mxGetPr( C_IN );
        complex = *Pr != 0;
    }
    else
        complex   = false;

    /* Create a matrix for the return arguments */
    if (complex)
    {
        Y_OUT = mxCreateNumericArray( nDims, dims, mxDOUBLE_CLASS, mxCOMPLEX );
        nCharToRead = 2*nToRead*nBytes;
    }
    else
    {
        Y_OUT = mxCreateNumericArray( nDims, dims, mxDOUBLE_CLASS, mxREAL );
        nCharToRead = nToRead*nBytes;
    }

    ch = mxMalloc( nCharToRead );

    ok = (*hp != 0);
    nRead = 0;

    /* Do it */
    if (ok & !ReadFile( *hp, ch, nCharToRead, &nRead, NULL))
        ok = FALSE; 

    /* move to output array */
    Pr  = mxGetPr( Y_OUT );

    if (complex)
    {
        Pi  = mxGetPi( Y_OUT );
        for (word_ind = 0, ch_ind   = 0; word_ind < nToRead; word_ind++)
        {
            word    = ch[ch_ind] << 24;
            ch_ind++;
            for (byte_ind = 1; byte_ind < nBytes; byte_ind++, ch_ind++)
                word = ((word >> 8) & 0xffffff) | (ch[ch_ind] << 24);
            word    >>= (sizeof( long ) - nBytes)*8;
            Pr[word_ind]    = word;

            word    = ch[ch_ind] << 24;
            ch_ind++;
            for (byte_ind = 1; byte_ind < nBytes; byte_ind++, ch_ind++)
                word = ((word >> 8) & 0xffffff) | (ch[ch_ind] << 24);
            word    >>= (sizeof( long ) - nBytes)*8;
            Pi[word_ind]    = word;
       }
    }
    else
    {
        for (word_ind = 0, ch_ind   = 0; word_ind < nToRead; word_ind++)
        {
            word    = ch[ch_ind] << 24;
            ch_ind++;
            for (byte_ind = 1; byte_ind < nBytes; byte_ind++, ch_ind++)
                word = ((word >> 8) & 0xffffff) | (ch[ch_ind] << 24);
            word    >>= (sizeof( long ) - nBytes)*8;
            Pr[word_ind]    = word;
        }
    }
            
    if (nlhs == 2)
    {
        STATUS = mxCreateDoubleMatrix(1, 1, mxREAL);
        Stat  = (double *)mxGetPr(STATUS);
        if (ok)
            *Stat = 1;
        else
            *Stat = 0;
    }

    mxFree( dims );
    mxFree( ch );
    
    return;
}

