/*
 *
 * cportOpen.C
 *
 * The calling syntax is:
 *
 *        H=cportopen(Port,Permission)
 *
 * Permission (optional) indicates that a file is to be opened,
 * it is one of the strings (compatible with fopen):
 *        'r'     read
 *        'w'     write (create if necessary)
 *        'a'     append (create if necessary)
 *        'r+'    read and write (do not create)
 *        'w+'    truncate or create for read and write
 *        'a+'    read and append (create if necessary)
 *        'W'     write without automatic flushing
 *        'A'     append without automatic flushing
 *
 */

#include <mex.h>
#include "mxadd.h"                                              /* added because mxArrayToString is not present in Matlab 5.1 */
#include <stdio.h>
#include <windows.h>

/* Input Arguments */

#define    PORT_IN      prhs[0]
#define PERMISSION_IN   prhs[1]

/* Output Arguments */

#define    HPort        plhs[0]

BOOL CPortOpen(
           HANDLE *hp,
           char *Port,
           char *Permission
           )
{
    BOOL ok;

    ok = TRUE;

    if (Permission != NULL)
    {
        if (strstr( Permission, "r+" ))
            *hp = CreateFile(Port, 
                  GENERIC_READ | GENERIC_WRITE,
                  0,
                  NULL,
                  OPEN_EXISTING,
                  0,
                  NULL);
        else if (strstr( Permission, "w+" ))
            *hp = CreateFile(Port, 
                  GENERIC_READ | GENERIC_WRITE,
                  0,
                  NULL,
                  CREATE_ALWAYS,
                  0,
                  NULL);
        else if (strstr( Permission, "a+" ))
            *hp = CreateFile(Port, 
                  GENERIC_READ | GENERIC_WRITE,
                  0,
                  NULL,
                  OPEN_ALWAYS,
                  0,
                  NULL);
        else if (strchr( Permission, 'r' ))
            *hp = CreateFile(Port, 
                  GENERIC_READ | GENERIC_WRITE,
                  0,
                  NULL,
                  OPEN_EXISTING,
                  0,
                  NULL);
        else if (strchr( Permission, 'w' ))
            *hp = CreateFile(Port, 
                  GENERIC_WRITE,
                  0,
                  NULL,
                  CREATE_ALWAYS,
                  0,
                  NULL);
        else if (strchr( Permission, 'a' ))
            *hp = CreateFile(Port, 
                  GENERIC_WRITE,
                  0,
                  NULL,
                  OPEN_ALWAYS,
                  0,
                  NULL);
        else
            mexErrMsgTxt("CPORTOPEN requires valid PERMISSION string.");

        if (*hp == INVALID_HANDLE_VALUE)
            ok = FALSE;
    }
    else
    {
        *hp = CreateFile(Port, 
                  GENERIC_READ | GENERIC_WRITE,
                  0,
                  NULL,
                  OPEN_EXISTING,
                  0,
                  NULL);
        if (*hp == INVALID_HANDLE_VALUE)
            ok = FALSE;
  
        if (ok)
            PurgeComm(*hp, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
    }

    if (ok)
        return TRUE;
    else
    {
        *hp = 0;
        return FALSE;
    }
}


void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
         )
{
    char        *Port, *Permission = NULL,
                PortStr[6], PortStrFull[_MAX_PATH], PortTest[6];
    const char  *ErrStr = "CPORTOPEN illegal port name.";
    HANDLE      hp[4], *PHandle;

    /* Check for proper number of arguments */
    if (nrhs < 1)
        mexErrMsgTxt("CPORTOPEN requires at least one input argument.");
    
    if (nlhs != 1)
        mexErrMsgTxt("CPORTOPEN requires one output argument.");
  
    /* Check the type of PORT_IN - should be char array */
    if (!mxIsChar(PORT_IN))
        mexErrMsgTxt("CPORTOPEN first argument should be a string.");

    /* Allocate and translate the input parameter string */
    Port   = mxArrayToString(PORT_IN);

    strcpy(PortStr,Port);
    mxFree(Port);

    if (nrhs == 1)
    {
        PortStr[5] = '\0';

        /* Test to see if asking for a valid com port */
        strcpy( PortTest, PortStr );
        PortTest[3] = '\0';
        if (strcmp(PortTest,"com")==0)
        {
            if ((PortStr[3]<'1') | (PortStr[3]>'8'))
                mexErrMsgTxt(ErrStr);
        }
        else if (strcmp(PortTest,"lpt")==0)
        {
            if ((PortStr[3]<'1') | (PortStr[3]>'3'))
                mexErrMsgTxt(ErrStr);
        }
        else
            mexErrMsgTxt(ErrStr);

        strcpy(PortStrFull, "\\\\.\\");    /* use the full legal device name */
        strcat(PortStrFull, PortStr);
    }
    else
    {
        /* Check the type of PERMISSION_IN - should be char array */
        if (!mxIsChar( PERMISSION_IN ))
            mexErrMsgTxt("CPORTOPEN second argument should be a string.");

        /* Allocate and translate the input parameter string */
        Permission   = mxArrayToString( PERMISSION_IN );

        strcpy( PortStrFull, Port );
    }
  
    /* Create a matrix for the return arguments and assign pointers */
    HPort = mxCreateDoubleMatrix(1, 1, mxREAL);
    PHandle = (HANDLE *)mxGetData(HPort);

    /* Do the actual operations in a subroutine */
    if (CPortOpen(hp, PortStrFull, Permission))
    {
        hp[1] = NULL;
        hp[2] = NULL;
        hp[3] = NULL;
    }
    else
    {
        hp[0] = NULL;
        hp[1] = NULL;
        hp[2] = NULL;
        hp[3] = NULL;
    } 

    *PHandle = *hp;

    return;
}

