/*=======================================================================+
|             DEPARTAMENTO MECANICA UNIVERSIDADE AVEIRO                  |
|               Projecto Humanoide (PHUA - MODULROB)                     |
|                                                                        |
| Programadores:                                                         |
|              David Manuel Costa Gameiro          N:20281               |
|              Filipe Carvalho Mostardinha         N:22085               |
|                                                                        |
| Nome Ficheiro: MasterSec.c                                             |
|                                                                        |
|                            SOURCE FILE                                 |
+=======================================================================*/

#include "usart.h"
#include "can.h"
#include "pic.h"

//////////////////////////////////////////
//   Mensagens tipo na recepcao         //
//////////////////////////////////////////

// _____ PACOTE DE TIPO DAS MENSAGENS VINDAS DO PC _______________________________________________________
// Mscaras para leitura dos diversos campos do pacote de tipo
enum MESSAGE_TYPE_PACKET  {
    // Campos do pacote de tipo
    OPERATION_MASK = 0b11000000,     // 0b11xxxxxx: 2 bits indicando o tipo de msg
    SCUID_MASK     = 0b00111100,     // 0bxx1111xx: 4 bits para a identificao do SCU
    PARAM_MASK     = 0b00000011,     // 0bxxxxxx11: 2 bits para a identificao do parmetro a ler/aplicar

    // Localizao dos vrios campos no pacote de tipo
    OPERATION_LOC  = 6,              // 0b11XXXXXX
    SCUID_LOC      = 2,              // 0bXX1111XX
    PARAM_LOC      = 0,              // 0bXXXXXX11

    // Tipos de operaes (identificado pelos bits 0b11xxxxxx)
    OP_APPLY_JOINT  = 0b00,           // Aplicao de novas posies/velocidades/correntes a uma junta
    OP_APPLY_CONTROL= 0b01,           // Aplicao de novos parametros a um controlador
    OP_READ_SENSORS = 0b10,           // Leitura dos sensores adicionais
    OP_READ_EXTBUFF = 0b11,           // Leitura do buffer externo

    // Tipos de parmetros (identificado pelos bits 0bxxxxxx11) a utilizar nas mensagens OP_APPLY_JOIND e OP_READ_JOINT
    PARAM_POSITION  = 0b00,           // Posio
    PARAM_VELOCITY  = 0b01,           // Velocidade
    PARAM_CURRENT   = 0b10,           // Corrente
    PARAM_SPECIAL   = 0b11,           // Sensores especiais

    // Parmetros do controlador (identificado pelos bits 0bxxxxxx11) a utilizar nas mensagens OP_APPLY_CONTROL
    PARAM_KI        = 0b00,
    PARAM_KP        = 0b01,
    PARAM_KD        = 0b10,
    PARAM_CONTROLON = 0b11,
};

// Macros para leitura dos diversos campos no pacote de tipo
#define operation(x) (((byte) (x) & OPERATION_MASK) >> OPERATION_LOC)   // Tipo de mensagem
#define scuID(x)     (((byte) (x) & SCUID_MASK    ) >> SCUID_LOC    )   // Identificao do SCU
#define param(x)     (((byte) (x) & PARAM_MASK    ) >> PARAM_LOC    )   // Identificao do parametro

// Macros para construo dos diversos padres no pacote de tipo
#define operation_patt(y)   (byte) (((y) << OPERATION_LOC) & OPERATION_MASK)   // Tipo de mensagem
#define scuID_patt(y)       (byte) (((y) << SCUID_LOC    ) & SCUID_MASK    )   // Identificao do SCU
#define param_patt(y)       (byte) (((y) << PARAM_LOC    ) & PARAM_MASK    )   // Identificao do parametro

// Macro para construo do pacote de tipo da mensagem
#define typePacketSensors(op,scu_id,param) (operation_patt(op) | scuID_patt(scu_id) | param_patt(param))
// ___________________________________________________________________________________________________________

// Buffer de software da USART
volatile static struct  {
    byte rx[USART_RECV_LENGTH]; // Buffer de recepo
    byte tx[USART_SEND_LENGTH]; // Buffer de transmisso
    byte tx_i, rx_i;            // ndices de acesso aos buffers
} usartBuffer;

// Buffer de segundo nvel, apelidado por externo
volatile static struct  {
    byte tx[USART_EXTBUFF_LENGTH];
    bool read;
} extBuffer;

// ****************************************************************************
// Configurao do Timer0 para definio do timeout para recepo e tranmisso de sequncias de bytes
#define TIMERO_PRES_CODE    0b101
#define TIMER0_PRES 64
#define TIMER0_OUT  100         // Timeout definido para 100ms
#define TIMER0L_INIT    (byte) (timer0_init & 0xFF)
#define TIMER0H_INIT    (byte) (timer0_init >> 8)
static word timer0_init=((word) 65535-(word) F_CPU*1000/(word) TIMER0_PRES*(word) TIMER0_OUT+1+2);


//*****************************************************************************
static void initUsartBuffers(void);                     // Inicializao dos buffers da USART

// Inicializacao da USART
//
void usartInit()
{
	float spbrg;
	
    // Inicializao dos buffers (de software) da USART
    initUsartBuffers();

    // Programacao dos pinos RC7 e RC6 como entrada e sada respectivamente
	TRISC=(TRISC | 0x80) & 0xBF;
	
    // Clculo do valor associado ao baudrate pedido
	spbrg=((float) F_CPU*1e6)/((float) 4*BAUD)-1.0;
	if (spbrg-(unsigned char) spbrg>=0.5)  spbrg+=1.0;

    // Programacao da Usart (Assincrona por Polling)
	SPBRG=(unsigned char) spbrg;    // BaudRate
	TXSTAbits.SYNC = 0;             // Modo assincrono
    TXSTAbits.BRGH = 1;	            // BR high speed
	TXSTAbits.TXEN = 1;	            // Transmit Enable 
    RCSTAbits.CREN = 1;	            // Continuous Reception Enable
	RCSTAbits.SPEN = 1;	            // Serial Port Enable

    IPR1bits.RCIP=0;                // Interrupes de baixa prioridade
    IPR1bits.TXIP=0;
    PIE1bits.RCIE=0;                // (Des)Activao das interrupes
    PIE1bits.TXIE=0;

    // Programao do Timer0 para deteco de timeouts
    T0CONbits.T08BIT=0;             // Timer em modo de 16 bits
    T0CONbits.T0CS=0;               // Internal instruction cycle clock (CLKO)
    T0CONbits.PSA=0;                // Timer0 prescaler is assigned. Timer0 clock input comes from prescaler output

    T0CONbits.T0PS2=(TIMERO_PRES_CODE>>2) & 1;      // Prescaler
    T0CONbits.T0PS1=(TIMERO_PRES_CODE>>1) & 1;
    T0CONbits.T0PS0=TIMERO_PRES_CODE & 1;

    //usartCountTimeStart();          // Inicializao do valor inicial do Timer0
    INTCONbits.TMR0IE=0;            // Interrupes desactivadas
    T0CONbits.TMR0ON=1;             // Enables Timer0
}

void usartCountTimeStart(void)
{
    TMR0H=TIMER0H_INIT;             // Valor inicial do timer0
    TMR0L=TIMER0L_INIT;
 
    INTCONbits.TMR0IF=0;
}

bool usartCountTimeout(void)
{
    return INTCONbits.TMR0IF;
}

//*****************************************************************************
// Inicializao dos buffers da Usart
//
static void initUsartBuffers(void)
{
    byte i;

    // Inicializao a zero de toda a estrutura usartBuffer
    for (i=0; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
    for (i=0; i<USART_RECV_LENGTH; ++i)  usartBuffer.rx[i]=0;

    usartBuffer.tx_i=usartBuffer.rx_i=0;

    // Inicializao do buffer externo
    for (i=0; i<USART_EXTBUFF_LENGTH; ++i)  extBuffer.tx[i]=0;
    extBuffer.read=true;
}

// _____ FUNES DE MANIPULAO DIRECTA DO BUFFER RX/TX __________________

//*****************************************************************************
// Armazenamento de uma mensagem no buffer externo da USART.
// Este buffer  necessrio para a transmisso uma mensagem
// configurvel pelo utilizador, seguindo a poltica Request->Send!
//
void usartStore2extBuff(byte *message, byte n)
{
    byte i;

    // Caso a mensagem apresentada seja demasiado grande, trunc-la.
    if (n>USART_EXTBUFF_LENGTH)  n=USART_EXTBUFF_LENGTH;

    // Preenchimento do buffer externo
    for (i=0; i<n; ++i)  extBuffer.tx[i]=message[i];

    // Finalizao do preenchimento com zeros
    while (i<USART_EXTBUFF_LENGTH)  extBuffer.tx[i++]=0;

    // A flag indicadora de que o buffer j foi acessado  inicializada a falso!
    extBuffer.read=false;
}

//*****************************************************************************
// Retorna true, se o buffer de recepo est a receber um stream de bytes
//
bool usartStreamMode(void)
{
    return (usartBuffer.rx_i>0);
}

//*****************************************************************************
// Insero de um byte no buffer de recepo.
// Retorna true, se a insero teve sucesso
// *full indica se o buffer de recepo est cheio
//
bool usartStoreRx(byte val, bool *full)
{
    // O buffer rx est cheio?
    if (usartBuffer.rx_i>=USART_RECV_LENGTH)  {
        if (full!=(bool *) 0)  *full=true;      // Buffer cheio!
        return false;                           // Insero falhada
    }

    // Insero do novo byte
    usartBuffer.rx[usartBuffer.rx_i++]=val;

    // Verificao de buffer cheio
    if (full!=(bool *) 0)  {
        if (usartBuffer.rx_i==USART_RECV_LENGTH)  *full=true;
        else                                      *full=false;
    }
    // Insero foi um sucesso
    return true;
}

//*****************************************************************************
// Obteno e remoo de um byte do buffer de transmisso.
// *var conter o byte removido;
// *empty indica se o buffer de transmisso est vazio
// True  retornado se a obteno teve sucesso.
//
bool usartGetTx(byte *var, bool *empty)
{
    // Verificando se o buffer de transmisso est vazio
    if (usartBuffer.tx_i>=USART_SEND_LENGTH)  {
        if (empty!=(bool *) 0)  *empty=true;    // Buffer de transmisso vazio
        return false;                           // Obteno do byte no efectuada
    }

    // Obteno do byte
    if (var!=(byte *) 0)  *var=usartBuffer.tx[usartBuffer.tx_i++];

    // Verificando se o buffer de transmisso ficou vazio
    if (empty!=(bool *) 0)  {
        if (usartBuffer.rx_i==USART_RECV_LENGTH)  *empty=true;
        else                                      *empty=false;
    }
    // Obteno efectuada com sucesso
    return true;
}

//*****************************************************************************
// Reset do Buffer de recepo.
// O buffer  esvaziado e pronto para a recepo de novos caracteres.
//
void usartResetRxBuff(void)
{
    usartBuffer.rx_i=0;
}

//*****************************************************************************
// Reset do Buffer de transmisso.
// O buffer  esvaziado e pronto para armazenar novos bytes.
//
void usartResetTxBuff(void)
{
    usartBuffer.tx_i=0;
}

// _____ FUNES PARA PREPARAO DE MENSAGENS DE RESPOSTA ________________

//*****************************************************************************
// Envio de uma mensagem de teste (FA F9 F8 F7 F6 F5)
//
bool usartSendTestMsg(void)
{
    byte i;

    for (i=0; i<USART_SEND_LENGTH; ++i)
        usartBuffer.tx[i]=MESSAGE_TEST-i;

    return true;
}

//*****************************************************************************
// Envio de uma mensagem de status
//
bool usartSendStatus(byte stat)
{
    byte i;

    // O pacote SOF conter o status a enviar:
    usartBuffer.tx[SOF]=stat;

    // Preenchimento do resto da mensagem com o contedo da ltima recebida
    for (i=TYPE; i<USART_RECV_LENGTH; ++i)  usartBuffer.tx[i]=usartBuffer.rx[i];

    // Preenchimento com zeros do resto da mensagem
    for (i=USART_RECV_LENGTH; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;

    return true;
}

//*****************************************************************************
static void newJointInfo(byte scu_id, byte param);      // Actualizao dos dados de uma junta
static void newJointControl(byte scu_id, byte param);   // Actualizao dos parmetros do controlador de uma junta

static void sendSensorsInfo(byte scu_id, byte param);   // Leitura sensorial de um dos SCU's
static void sendJointInfo(byte scu_id, byte param);     // Envio das informaes sensoriais de uma junta
static void sendSpecialInfo(byte scu_id);               // Envio da matriz dos sensores especiais
static void sendExtBuffer(byte typePacket);             // Envio de uma mensagem com o contedo do buffer externo

// Processamento de uma mensagem recebida pela Usart (a partir do PC)
//
bool usartProcessMsg(void)
{
    byte msg_type;

    // Padro descritor do tipo da mensagem
    msg_type=usartBuffer.rx[TYPE];

    // De acordo com a operao referente  mensagem...
    switch (operation(msg_type))  {

        // Aplicao de uma nova posio, velocidade ou corrente a todos os servos de uma junta
        case OP_APPLY_JOINT:    newJointInfo(scuID(msg_type),param(msg_type));
                                break;

        // Aplicao do controlador
        case OP_APPLY_CONTROL:  newJointControl(scuID(msg_type),param(msg_type));
                                break;

        // Leitura de um parmetro de todos os servos de uma junta
        case OP_READ_SENSORS:   sendSensorsInfo(scuID(msg_type),param(msg_type));
                                break;

        // Leitura dos sensores adicionais (sensores de fora, inclinmetro ou giroscpio)
        case OP_READ_EXTBUFF:   sendExtBuffer(typePacketSensors(OP_READ_EXTBUFF,0,0));
                                break;
    }
    return true;
}


//*****************************************************************************
// Aplicao de uma nova posio, velocidade ou corrente a todos os servos de uma junta
//
static void newJointInfo(byte scu_id, byte param)
{
    byte scu, servo;

    // Validao do ndice do SCU
    if ((scu=scuIndex(scu_id))>=N_SCU)  {
        usartSendStatus(MESSAGE_INVREQ);
        return;
    }

    // Condies especiais de funcionamento do sistema
    if (param==PARAM_SPECIAL)  {
        actuators[scu].sysStatus.pwm=usartBuffer.rx[DATA1] & 1;
        actuators[scu].sysStatus.calib=usartBuffer.rx[DATA1+1] & 1;
    }
    // Actuao sobre o sistema
    else  {
        // Preenchimento dos arrays internos com os novos dados recebidos (posio, velocidade ou corrente)
        for (servo=0; servo<N_SERVOS; ++servo)
            switch (param)  {
                // Posio
                case PARAM_POSITION:    actuators[scu].servo[servo].position=usartBuffer.rx[DATA1+servo];
                                        break;
                // Velocidade
                case PARAM_VELOCITY:    actuators[scu].servo[servo].velocity=usartBuffer.rx[DATA1+servo];
                                        break;
                // Corrente
                case PARAM_CURRENT:     actuators[scu].servo[servo].current =usartBuffer.rx[DATA1+servo];
                                        break;
            }
    }

    // Envio de uma mensagem indicadora que a operao pedida teve sucesso!
    usartSendStatus(MESSAGE_SUCESS);
}

//*****************************************************************************
// Aplicao de uma nova posio, velocidade ou corrente a todos os servos de uma junta
//
static void newJointControl(byte scu_id, byte param)
{
    byte scu, servo;

    // Validao do ndice do SCU
    if ((scu=scuIndex(scu_id))>=N_SCU)  {
        usartSendStatus(MESSAGE_INVREQ);
        return;
    }

    // Preenchimento dos arrays internos com os novos dados recebidos (posio, velocidade ou corrente)
    for (servo=0; servo<N_SERVOS; ++servo)
        switch (param)  {
            // Parmetro proporcional
            case PARAM_KI:          actuators[scu].control[servo].ki=usartBuffer.rx[DATA1+servo];
                                    break;
            // Parmetro derivativo
            case PARAM_KP:          actuators[scu].control[servo].kp=usartBuffer.rx[DATA1+servo];
                                    break;
            // Parmetro integrativo
            case PARAM_KD:          actuators[scu].control[servo].kd=usartBuffer.rx[DATA1+servo];
                                    break;
            // Parmetro on/off dos controladores
            case PARAM_CONTROLON:   actuators[scu].control[servo].type=usartBuffer.rx[DATA1+servo] & 0b11;
                                    break;
        }

    // Envio de uma mensagem indicadora que a operao pedida teve sucesso!
    usartSendStatus(MESSAGE_SUCESS);
}


//*****************************************************************************
// Leitura sensorial de um dos SCU's
//
static void sendSensorsInfo(byte scu_id, byte param)
{
    // Validao do ndice do SCU
    if (scuIndex(scu_id)>=N_SCU)  {
        usartSendStatus(MESSAGE_INVREQ);
        return;
    }

    if (param==PARAM_SPECIAL)  sendSpecialInfo(scu_id);
    else                       sendJointInfo(scu_id,param);
}


//*****************************************************************************
// Leitura de um parmetro de todos os servos de uma junta
//
static void sendJointInfo(byte scu_id, byte param)
{
    byte i, servo, scu=scuIndex(scu_id);
    byte typePacket=typePacketSensors(OP_READ_SENSORS,scu_id,param);

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;                                // Start of Frame
    usartBuffer.tx[TYPE]=typePacket;                                    // Tipo da mensagem

    // Estado dos servos
    // Preenchimento do buffer de transmisso com as informaes sensoriais da junta
    for(servo=0; servo<N_SERVOS; ++servo)
        switch (param)  {
            // Posio
            case PARAM_POSITION:    usartBuffer.tx[DATA1+servo]=sensors[scu].servo[servo].position;
                                    break;
            // Velocidade
            case PARAM_VELOCITY:    usartBuffer.tx[DATA1+servo]=sensors[scu].servo[servo].velocity;
                                    break;
            // Corrente
            case PARAM_CURRENT:     usartBuffer.tx[DATA1+servo]=sensors[scu].servo[servo].current;
                                    break;
            // Outros
            default:                usartBuffer.tx[DATA1+servo]=0;
        }

    // Pacote de estado do sistema
    usartBuffer.tx[DATA1+N_SERVOS]=sensors[scu].sysStatus;

    // Preenchimento com zeros do resto da mensagem
    for (i=DATA1+N_SERVOS+1; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
//    PIE1bits.TXIE=1;
}


//*****************************************************************************
// Leitura da matriz dos sensores de fora de um determinado p
//
static void sendSpecialInfo(byte scu_id)
{
    byte i, scu=scuIndex(scu_id);

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;                                    // Start Of Frame
    usartBuffer.tx[TYPE]=typePacketSensors(OP_READ_SENSORS,scu_id,PARAM_SPECIAL);  // Tipo da mensagem

    // Preenchimento do buffer de transmisso com os valores dos sensores de fora
    for(i=0; i<N_SPECIAL_SENSORS; ++i)  usartBuffer.tx[DATA1+i]=sensors[scu].special[i];

    // Preenchimento com zeros do resto da mensagem
    for (i=DATA1+N_SPECIAL_SENSORS; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
//    PIE1bits.TXIE=1;
}

//*****************************************************************************
// Envio do contedo do buffer externo pela USART
//
static void sendExtBuffer(byte typePacket)
{
    byte i;

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;
    usartBuffer.tx[TYPE]=typePacket;

    // Cpia dos dados do buffer externo para o interno
    for (i=0; i<USART_EXTBUFF_LENGTH; ++i)  {
        usartBuffer.tx[DATA1+i]=extBuffer.tx[i];    // Transferncia da informao
        extBuffer.tx[i]=0;                          // Clear do buffer origem
    }

    // A flag read indica que os dados foram lidos pela USART
    extBuffer.read=true;

    // Reset do estado de erro do CAN
    canClearStatus();
}
