#include "servo.h"
#include "pic2.h"
#include "can.h"
//#include "com.h"
#include "data.h"

// Estrutura com os dados do controlador
typedef struct {
    double angle;               // ngulo a solicitar
    double deltaD_prev;         // Variao derivativa do ciclo anterior
    signed int  error_prev;     // Erro do ciclo anterior
    signed char y_prev;         // Sinal de feedback (potencimetro) do ciclo anterior
    signed char y_prev2;        // Sinal de feedback de h 2 ciclos atrs
} struct_controller;

// Estrutura com os dados do controlador
typedef struct  {
    signed char finalPosition;      // Posio final a atingir
    struct_controller controller;   // Dados do controlador
} struct_servoControl;
static struct_servoControl servoControl[N_SERVOS];      // Dados dos actuadores


// Execuo da calibrao dos servos
//
void calibration(void)
{
    byte i, count;
    signed int med;
    unsigned int special_sum[N_SPECIAL_SENSORS];
    signed int position_sum[N_SERVOS];
    signed char posReq[N_SERVOS];

    // Esperando por uma ordem do Master
    while (!canRequest);

    // ___ CALIBRAO DOS SENSORES DE FORA ____________________________
    wait(2000);

    // Inicializao das variveis locais
    for (i=0; i<N_SPECIAL_SENSORS; ++i)  special_sum[i]=0;

    // Execuo de 25 medidas
    for (count=0; count<25; ++count)  {
        // Esperando pela passagem de um tick;
        waitTick();

        // Acumulao da sada dos sensores de fora
        for (i=0; i<N_SPECIAL_SENSORS; ++i)
            special_sum[i] += sensors.special[i];
    }

    // Correco dos desvios dos sensores de fora
    for (i=0; i<N_SPECIAL_SENSORS; ++i)  {
        med = (signed int) (special_sum[i]/25);
        // Offset determinado
        special0[i] = (unsigned char) 128 - (unsigned char) med;
    }

    // ___ CALIBRAO DOS POTENCIMETROS DOS SERVOS ____________________
    // Inicializao de variveis relativas aos servos e activao do PWM
    for (i=0; i<N_SERVOS; ++i)  {

        // Posio solicitada
        posReq[i]=actuators.servo[i].position;

        // Definio da posio inicial
        servoActuation(posReq[i],i);

        // Inicializao do array
        position_sum[i]=0;
    }

    // Activao do sinal de PWM
    statusPWM(true);

    // Esperando 2 segundos para que a posio desejada seja atingida
    wait(2000);

    // Execuo de 25 medidas
    for (count=0; count<25; ++count)  {
        // Esperando pela passagem de um tick;
        waitTick();

        // Acumulao das posies medidas
        for (i=0; i<N_SERVOS; ++i)
            position_sum[i] += servo_interface[i].position;
    }

    // Correco dos desvios da medio do potencimetro
    for (i=0; i<N_SERVOS; ++i)  {
        med = (signed int) (position_sum[i]/25);
        origin[i] += (signed int) posReq[i]-med;
    }
    // Calibrao efectuada!
}

// Movimentao do servo para a posio solicitada
//
signed char servoRequest(signed char position, byte srv)
{
    // Correco da posio solicitada
    position=limitPosition(position);

    // Posio a solicitar
    servoControl[srv%N_SERVOS].finalPosition=position;

    return position;
}

// ==========================================================================================
// Funes temporais
// ==========================================================================================

// Inicializao do Controlador
//
void initController(void)
{
    byte srv;
    struct_controller *cont;

    for (srv=0; srv<N_SERVOS; ++srv)  {

        cont = &(servoControl[srv].controller);

        servoControl[srv].finalPosition=actuators.servo[srv].position;
        cont->angle=(double) servoControl[srv].finalPosition;
        cont->deltaD_prev=0.0;
        cont->error_prev=0;
        cont->y_prev=0;
        cont->y_prev2=0;
    }
}

static void openloopControl(byte servo);
static void locomotionControl(byte servo);
static void reactionControl(byte servo, signed int copx, signed int copy);
static void jacobianControl(byte servo, signed int copx, signed int copy, signed int Jt[N_SERVOS][N_SERVOS]);

void controller(void)
{
    byte servo;
    signed int copx, copy, Jt[N_SERVOS][N_SERVOS];

    // Clculo do centro de presso (em dcimas de mm)
    calcCoP(&copx,&copy);

    // Clculo da matriz Jacobiana
    jacobianMatrix(Jt);

    for (servo=0; servo<N_SERVOS; ++servo)  {

        // Correco do tipo de controlo a usar 
        if (actuators.control[servo].type==LOCOMOTION_CONTROL && actuators.control[servo].kp==0
            && actuators.control[servo].kd==0 && actuators.control[servo].ki==0)
            actuators.control[servo].type=NO_CONTROL;

        // Controlo em malha aberta
        openloopControl(servo);

        // Controlo de posio em malha fechada
        locomotionControl(servo);

        // Controlo das foras de reaco
        reactionControl(servo,copx,copy);

        // Controlo de equilbrio
        jacobianControl(servo,copx,copy,Jt);
    }
}

// Limitao do controlo de posio
static signed char limitControlPosition(byte servo);

static void openloopControl(byte servo)
{
    // Se o controlo em malha aberta no estiver contemplado, abandonar!
    if (actuators.control[servo].type!=NO_CONTROL)  return;

    // CONTROLO EM MALHA ABERTA
    // Passagem da posio solicitada  a plicar
    servoControl[servo].controller.angle=(double) servoControl[servo].finalPosition;
    
    // Aplicao do movimento
    servoActuation(limitControlPosition(servo),servo);
}


#define FILTER_N  1.0           // Nvel de filtragem da componente derivativa

static void locomotionControl(byte servo)
{
    signed char posReference, y_now;
    signed int error, error_dif, y_vardif;
    double deltaP, deltaI, deltaD, kp, ki, kd;
    struct_controller *cont;

    posReference=servoControl[servo].finalPosition;       // Posio desejada (referncia)
    y_now = servo_interface[servo].position;              // Posio efectiva do servo
    
    // Erro actual
    error=(signed int) posReference - (signed int) y_now;
    // banda morta
    if (error>=-2 && error<=2)  error= 0;
    else if (error== 3)         error= 2;
    else if (error==-3)         error=-2;
    
    // Diferena do erro para a aco proporcional e variao da diferena da sada para a aco derivativa
    error_dif= error-cont->error_prev;
    y_vardif =(signed int) y_now-2*(signed int) cont->y_prev+(signed int) cont->y_prev2;
    
    // Ponteiro para a estrutura de controlo do servo
    cont = &(servoControl[servo].controller);
    
    // Actualizao das sadas/erros anteriores
    cont->y_prev2=cont->y_prev;
    cont->y_prev =y_now;
    cont->error_prev=error;
    
    // Se o controlo de posio em malha fechada no for contemplado passar ao servo seguinte!
    if (actuators.control[servo].type!=LOCOMOTION_CONTROL)  return;
     
    // CONTROLO EM MALHA FECHADA DA LOCMOO
    
    ki = (double) actuators.control[servo].ki/(double) K_FACTOR;
    kp = (double) actuators.control[servo].kp/(double) K_FACTOR;
    kd = (double) actuators.control[servo].kd/(double) K_FACTOR;
    
    // Componentes do compensador
    deltaI = (double) error*ki;                     // Compensao integrativa
    deltaP = (double) error_dif*kp;                 // Compensao proporcional
    deltaD = (double) y_vardif*kd;                  // Compensao derivativa
    
    // Filtragem
    //            deltaD = (pwmServo[servo].control.deltaD_prev*kd/FILTER_N + deltaD)/(1+kd/FILTER_N);
    cont->deltaD_prev=deltaD;   // Update da ltima variao de derivada
    
    // Clculo da nova posio a aplicar
    cont->angle += deltaI + deltaP + deltaD;
    
    // Aplicao do movimento
    servoActuation(limitControlPosition(servo),servo);
}


// _____ CONTROLADOR DE EQUILBRIO PROPORCIONAL _______________________________________
// Processamento e aplicao da nova velocidade a aplicar
static void applyVelocity(byte servo, double velocity);

#define KP_FORCEY    5e-2
#define KP_FORCEX    1e-1

static void reactionControl(byte servo, signed int copx, signed int copy)
{
    double delta;

    // Se o tipo de controlo a aplicar no for de reaco, abandonar funo!
    if (actuators.control[servo].type!=REACTION_CONTROL)  return;

    if (servo==0)  {
        // COMPENSAO DO CENTRO DE PRESSO NO EIXO XX
        // Clculo da variao de posio a aplicar
        delta = (double) -copx*(double) KP_FORCEX;
    
        // Aplicao da nova velocidade ao servo 0
        applyVelocity(servo,delta);
    }
    else if (servo==1)  {
        // COMPENSAO DO CENTRO DE PRESSO NO EIXO YY
        // Actualizao da posio do servo
        delta = (double) copy*(double) KP_FORCEY;
    
        // Aplicao da nova velocidade ao servo 1
        applyVelocity(servo,delta);
    }
}


// Constantes de Ganho da matriz Jacobiana
static const double jacobKp[N_SERVOS]={-0.1, -0.1, 0.1};

// _____ CONTROLADOR DE EQUILBRIO PELO JACOBIANO _____________________________________
static void jacobianControl(byte servo, signed int copx, signed int copy, signed int Jt[N_SERVOS][N_SERVOS])
{
    double delta[N_SERVOS];

    // Se o tipo de controlo no for de equlbrio, sair!
    if (actuators.control[servo].type!=BALANCE_CONTROL)  return;

    // Clculo do incremento de posio
    delta[servo] = jacobKp[servo]*( (double) Jt[servo][0]*copx + (double) Jt[servo][1]*copy )/1e4;

    // Aplicao do incremento
    applyVelocity(servo,delta[servo]);
}


#define MAX_SPEED   50                              // Mxima velocidade (graus/s)
#define MAX_POSVAR  ((double) (MAX_SPEED)*(double) TICK_TIME/1000.0)

// Processamento e aplicao da nova velocidade a aplicar
static void applyVelocity(byte servo, double velocity)
{
    // Limitao de velocidade
    if (velocity<-MAX_POSVAR)      velocity = -MAX_POSVAR;
    else if (velocity>MAX_POSVAR)  velocity =  MAX_POSVAR;

    // Actualizao da posio do servo
    servoControl[servo].controller.angle += velocity;

    // Aplicao da nova posio
    servoActuation(limitControlPosition(servo),servo);
}


// Limitao da posio a aplicar  sada do controlador.
// O servo a limitar deve ser fornecido como parmetro
static signed char limitControlPosition(byte servo)
{
    struct_controller *cont;

    // Ponteiro para a estrutura de controlo correspondente
    cont = &(servoControl[servo].controller);

    // Limitao da posio a solicitar
    if ((signed int) cont->angle<-SERVO_RANGE)      cont->angle=-(double) SERVO_RANGE;
    else if ((signed int) cont->angle>SERVO_RANGE)  cont->angle= (double) SERVO_RANGE;
    
    return ((signed char) cont->angle);
}
