/*=======================================================================+
|             DEPARTAMENTO MECANICA UNIVERSIDADE AVEIRO                  |
|               Projecto Humanoide (PHUA - MODULROB)                     |
|                                                                        |
| Programadores:                                                         |
|              David Manuel Costa Gameiro          N:20281               |
|              Filipe Carvalho Mostardinha         N:22085               |
|                                                                        |
| Nome Ficheiro: MasterSec.c                                             |
|                                                                        |
|                            SOURCE FILE                                 |
+=======================================================================*/

#include "usart.h"
#include "can.h"
#include "pic.h"

// Baud Rate das comunicaes pela USART
#define BAUD	115200

// ndices dos pacotes de uma mensagem
#define SOF         0                   // Start of Frame
#define TYPE        1                   // Tipo da mensagem
#define DATA1       2                   // Primeiro dado

// Comprimento em bytes das mensagens de/para o PC
#define USART_SEND_LENGTH    (DATA1+N_SPECIAL_SENSORS)   // Mensagens PIC -> PC  (SOF+Type+NForceSensors)
#define USART_RECV_LENGTH    (DATA1+N_SERVOS)            // Mensagens PC  -> PIC (SOF+Type+NServos)
#define USART_EXTBUFF_LENGTH (USART_SEND_LENGTH-DATA1)   // Buffer externo para transmisso (PIC -> PC)


//////////////////////////////////////////
//   Mensagens tipo na recepcao         //
//////////////////////////////////////////

// _____ PACOTE DE TIPO DAS MENSAGENS VINDAS DO PC _______________________________________________________
// Mscaras para leitura dos diversos campos do pacote de tipo
enum MESSAGE_TYPE_PACKET  {
    // Campos do pacote de tipo
    OPERATION_MASK = 0b11000000,     // 0b11xxxxxx: 2 bits indicando o tipo de msg
    SCUID_MASK     = 0b00111100,     // 0bxx1111xx: 4 bits para a identificao do SCU
    PARAM_MASK     = 0b00000011,     // 0bxxxxxx11: 2 bits para a identificao do parmetro a ler/aplicar

    // Localizao dos vrios campos no pacote de tipo
    OPERATION_LOC  = 6,              // 0b11XXXXXX
    SCUID_LOC      = 2,              // 0bXX1111XX
    PARAM_LOC      = 0,              // 0bXXXXXX11

    // Tipos de operaes (identificado pelos bits 0b11xxxxxx)
    OP_APPLY_JOINT  = 0b00,           // Aplicao de novas posies/velocidades/correntes a uma junta
    OP_APPLY_CONTROL= 0b01,           // Aplicao de novos parametros a um controlador
    OP_READ_SENSORS = 0b10,           // Leitura dos sensores adicionais
    OP_READ_EXTBUFF = 0b11,           // Leitura do buffer externo

    // Tipos de parmetros (identificado pelos bits 0bxxxxxx11) a utilizar nas mensagens OP_APPLY_JOIND e OP_READ_JOINT
    PARAM_POSITION  = 0b00,           // Posio
    PARAM_VELOCITY  = 0b01,           // Velocidade
    PARAM_CURRENT   = 0b10,           // Corrente
    PARAM_SPECIAL   = 0b11,           // Sensores especiais

    // Parmetros do controlador (identificado pelos bits 0bxxxxxx11) a utilizar nas mensagens OP_APPLY_CONTROL
    PARAM_KI        = 0b00,
    PARAM_KP        = 0b01,
    PARAM_KD        = 0b10,
    PARAM_CONTROLON = 0b11
};

// Macros para leitura dos diversos campos no pacote de tipo
#define operation(x) (((byte) (x) & OPERATION_MASK) >> OPERATION_LOC)   // Tipo de mensagem
#define scuID(x)     (((byte) (x) & SCUID_MASK    ) >> SCUID_LOC    )   // Identificao do SCU
#define param(x)     (((byte) (x) & PARAM_MASK    ) >> PARAM_LOC    )   // Identificao do parametro

// Macros para construo dos diversos padres no pacote de tipo
#define operation_patt(y)   (((byte) (y) << OPERATION_LOC) & OPERATION_MASK)   // Tipo de mensagem
#define scuID_patt(y)       (((byte) (y) << SCUID_LOC    ) & SCUID_MASK    )   // Identificao do SCU
#define param_patt(y)       (((byte) (y) << PARAM_LOC    ) & PARAM_MASK    )   // Identificao do parametro

// Macro para construo do pacote de tipo da mensagem
#define typePacketSensors(op,scu_id,param) (operation_patt(op) | scuID_patt(scu_id) | param_patt(param))
// ___________________________________________________________________________________________________________

////////////////////////////////////////////////
//   Buffers para armazenamento das mensagens //
////////////////////////////////////////////////

// Buffer de software da USART
static struct  {
    byte rx[USART_RECV_LENGTH+1]; // Buffer de recepo
    byte tx[USART_SEND_LENGTH+1]; // Buffer de transmisso
    byte tx_i, rx_i;            // ndices de acesso aos buffers
} usartBuffer;

// Buffer de segundo nvel, apelidado por externo
static struct  {
    byte tx[USART_EXTBUFF_LENGTH+1];
    bool read;
} extBuffer;

////////////////////////////////////////////////////////
//   Timer para definio do timeout nas comunicaes //
////////////////////////////////////////////////////////

// ****************************************************************************
// Configurao do Timer0 para definio do timeout para recepo e tranmisso de sequncias de bytes
#define TIMERO_PRES_CODE    0b101
#define TIMER0_PRES 64
#define TIMER0_OUT  100         // Timeout definido para 100ms
#define TIMER0L_INIT    (byte) (timer0_init & 0xFF)
#define TIMER0H_INIT    (byte) (timer0_init >> 8)
static word timer0_init=((word) 65535-(word) F_CPU*1000/(word) TIMER0_PRES*(word) TIMER0_OUT+1+2);

//////////////////////////////////
//   Rotinas de inicializao   //
//////////////////////////////////

//*****************************************************************************
static void initUsartBuffers(void);                     // Inicializao dos buffers da USART

// Inicializacao da USART
//
void usartInit()
{
	float spbrg;
	
    // Inicializao dos buffers (de software) da USART
    initUsartBuffers();

    // Programacao dos pinos RC7 e RC6 como entrada e sada respectivamente
	TRISC=(TRISC | 0x80) & 0xBF;
	
    // Clculo do valor associado ao baudrate pedido
	spbrg=((float) F_CPU*1e6)/((float) 4*BAUD)-1.0;
	if (spbrg-(unsigned char) spbrg>=0.5)  spbrg+=1.0;

    // Programacao da Usart (Assincrona por Polling)
	SPBRG=(unsigned char) spbrg;    // BaudRate
	TXSTAbits.SYNC = 0;             // Modo assincrono
    TXSTAbits.BRGH = 1;	            // BR high speed
	TXSTAbits.TXEN = 1;	            // Transmit Enable 
    RCSTAbits.CREN = 1;	            // Continuous Reception Enable
	RCSTAbits.SPEN = 1;	            // Serial Port Enable

    IPR1bits.RCIP=1;                // Alta prioridade
    PIE1bits.RCIE=1;                // (Des)Activao das interrupes
    PIE1bits.TXIE=0;
    PIR1bits.RCIF=0;

    // Programao do Timer0 para deteco de timeouts
    T0CONbits.T08BIT=0;             // Timer em modo de 16 bits
    T0CONbits.T0CS=0;               // Internal instruction cycle clock (CLKO)
    T0CONbits.PSA=0;                // Timer0 prescaler is assigned. Timer0 clock input comes from prescaler output

    T0CONbits.T0PS2=(TIMERO_PRES_CODE>>2) & 1;      // Prescaler
    T0CONbits.T0PS1=(TIMERO_PRES_CODE>>1) & 1;
    T0CONbits.T0PS0=TIMERO_PRES_CODE & 1;

    //usartCountTimeStart();          // Inicializao do valor inicial do Timer0
    INTCONbits.TMR0IE=0;            // Interrupes desactivadas
    T0CONbits.TMR0ON=1;             // Enables Timer0
}

//*****************************************************************************
// Inicializao dos buffers da Usart
//
static void initUsartBuffers(void)
{
    byte i;

    // Inicializao a zero de toda a estrutura usartBuffer
    for (i=0; i<USART_SEND_LENGTH+1; ++i)  usartBuffer.tx[i]=0;
    for (i=0; i<USART_RECV_LENGTH+1; ++i)  usartBuffer.rx[i]=0;

    usartBuffer.tx_i=usartBuffer.rx_i=0;

    // Inicializao do buffer externo
    for (i=0; i<USART_EXTBUFF_LENGTH+1; ++i)  extBuffer.tx[i]=0;
    extBuffer.read=true;
}

///////////////////////////////////////////////////////////////////
//   Rotinas de manipulao do timer para avaliao de timeout   //
///////////////////////////////////////////////////////////////////

void usartCountTimeStart(void)
{
    TMR0H=TIMER0H_INIT;             // Valor inicial do timer0
    TMR0L=TIMER0L_INIT;
 
    INTCONbits.TMR0IF=0;
}

bool usartCountTimeout(void)
{
    return INTCONbits.TMR0IF;
}

////////////////////////////////////////////////////////
//   FUNES DE MANIPULAO DIRECTA DO BUFFER RX/TX   //
////////////////////////////////////////////////////////

//*****************************************************************************
// Armazenamento de uma mensagem no buffer externo da USART.
// Este buffer  necessrio para a transmisso uma mensagem
// configurvel pelo utilizador, seguindo a poltica Request->Send!
//
void usartStore2extBuff(byte *message, byte n)
{
    byte i;

    // Caso a mensagem apresentada seja demasiado grande, trunc-la.
    if (n>USART_EXTBUFF_LENGTH)  n=USART_EXTBUFF_LENGTH;

    // Preenchimento do buffer externo
    for (i=0; i<n; ++i)  extBuffer.tx[i]=message[i];

    // Finalizao do preenchimento com zeros
    while (i<USART_EXTBUFF_LENGTH)  extBuffer.tx[i++]=0;

    // A flag indicadora de que o buffer j foi acessado  inicializada a falso!
    extBuffer.read=false;
}

//*****************************************************************************
// Retorna true, se o buffer de recepo est a receber um stream de bytes
//
bool usartStreamMode(void)
{
    return (usartBuffer.rx_i>0);
}

//*****************************************************************************
// Insero de um byte no buffer de recepo.
// Retorna true, se a insero teve sucesso
// *full indica se o buffer de recepo est cheio
//
bool usartStoreRx(byte val, bool *full)
{
    // O buffer rx est cheio? (o byte BCC  considerado)
    if (usartBuffer.rx_i>=USART_RECV_LENGTH+1)  {
        if (full!=(bool *) 0)  *full=true;      // Buffer cheio!
        return false;                           // Insero falhada
    }

    // Insero do novo byte
    usartBuffer.rx[usartBuffer.rx_i++]=val;

    // Verificao de buffer cheio
    if (full!=(bool *) 0)  {
        if (usartBuffer.rx_i==USART_RECV_LENGTH+1)  *full=true;
        else                                        *full=false;
    }
    // Insero foi um sucesso
    return true;
}

//*****************************************************************************
// Obteno e remoo de um byte do buffer de transmisso.
// *var conter o byte removido;
// *empty indica se o buffer de transmisso est vazio
// True  retornado se a obteno teve sucesso.
//
bool usartGetTx(byte *var, bool *empty)
{
    // Verificando se o buffer de transmisso est vazio (o byte BCC  considerado)
    if (usartBuffer.tx_i>=USART_SEND_LENGTH+1)  {
        if (empty!=(bool *) 0)  *empty=true;    // Buffer de transmisso vazio
        return false;                           // Obteno do byte no efectuada
    }

    // Obteno do byte
    if (var!=(byte *) 0)  *var=usartBuffer.tx[usartBuffer.tx_i++];

    // Verificando se o buffer de transmisso ficou vazio
    if (empty!=(bool *) 0)  {
        if (usartBuffer.rx_i==USART_RECV_LENGTH+1)  *empty=true;
        else                                        *empty=false;
    }
    // Obteno efectuada com sucesso
    return true;
}

//*****************************************************************************
// Reset do Buffer de recepo.
// O buffer  esvaziado e pronto para a recepo de novos caracteres.
//
void usartResetRxBuff(void)
{
    usartBuffer.rx_i=0;
}

//*****************************************************************************
// Reset do Buffer de transmisso.
// O buffer  esvaziado e pronto para armazenar novos bytes.
//
void usartResetTxBuff(void)
{
    usartBuffer.tx_i=0;
}

//////////////////////////////////////////////////////////
//   FUNES PARA PREPARAO DE MENSAGENS DE RESPOSTA   //
//////////////////////////////////////////////////////////

static void usartSendTestMsg(void);         // Envio de uma mensagem de resposta
static void usartSystemMsg(void);           // Envio de uma mensagem de resposta com a descrio do sistema

static void buildStatusAnswer(byte stat);     // Envio de uma mensagem de status
static void addBCC(void);                   // Adiciona o cdigo BCC ao fim de uma mensagem de transmisso
static bool checkBCC(void);                 // Verifica se o cdigo BCC presente numa mensagem recebida est correcto

//*****************************************************************************
// Envio de uma mensagem de status
//
void usartSendStatus(byte stat)
{
    // Construo da mensagem de status
    buildStatusAnswer(stat);

    // Adio do byte BCC
    addBCC();
}

// Funo principal para processamento das mensagens de resposta
void usartProcessMsg(void)
{
    byte msgtype=usartBuffer.rx[SOF];

    // Validao do byte BCC na mensagem recebida
    if (!checkBCC())
        buildStatusAnswer(MESSAGE_INVALID);
    else  {
        // Executar a funo correspondente  tarefa
        switch (msgtype)  {
        case MESSAGE_REQ:   usartSystemMsg();
                            break;
        case MESSAGE_TEST:  usartSendTestMsg();
                            break;
        }
    }

    // adicionar byte BCC
    addBCC();
}


//*****************************************************************************
// Envio de uma mensagem de teste
//
static void usartSendTestMsg(void)
{
    byte i;

    // Copiando o conteudo da mensagem de recepo para a de transmisso
    for (i=0; i<USART_RECV_LENGTH && i<USART_SEND_LENGTH; ++i)
        usartBuffer.tx[i]=usartBuffer.rx[i];
    // Preenchendo o resto com zeros
    for (; i<USART_SEND_LENGTH; ++i)
        usartBuffer.tx[i]=0;
}


//*****************************************************************************
static void newJointInfo(byte scu_id, byte param);      // Actualizao dos dados de uma junta
static void newJointControl(byte scu_id, byte param);   // Actualizao dos parmetros do controlador de uma junta

static void sendSensorsInfo(byte scu_id, byte param);   // Leitura sensorial de um dos SCU's
static void sendJointInfo(byte scu_id, byte param);     // Envio das informaes sensoriais de uma junta
static void sendSpecialInfo(byte scu_id);               // Envio da matriz dos sensores especiais
static void sendExtBuffer(byte typePacket);             // Envio de uma mensagem com o contedo do buffer externo

// Processamento de uma mensagem recebida pela Usart (a partir do PC)
//
static void usartSystemMsg(void)
{
    byte msg_type;

    // Padro descritor do tipo da mensagem
    msg_type=usartBuffer.rx[TYPE];

    // De acordo com a operao referente  mensagem...
    switch (operation(msg_type))  {

        // Aplicao de uma nova posio, velocidade ou corrente a todos os servos de uma junta
        case OP_APPLY_JOINT:    newJointInfo(scuID(msg_type),param(msg_type));
                                break;

        // Aplicao do controlador
        case OP_APPLY_CONTROL:  newJointControl(scuID(msg_type),param(msg_type));
                                break;

        // Leitura de um parmetro de todos os servos de uma junta
        case OP_READ_SENSORS:   sendSensorsInfo(scuID(msg_type),param(msg_type));
                                break;

        // Leitura dos sensores adicionais (sensores de fora, inclinmetro ou giroscpio)
        case OP_READ_EXTBUFF:   sendExtBuffer(typePacketSensors(OP_READ_EXTBUFF,0,0));
                                break;
    }
}


//*****************************************************************************
// Aplicao de uma nova posio, velocidade ou corrente a todos os servos de uma junta
//
static void newJointInfo(byte scu_id, byte param)
{
    byte scu, srv;
    far signed char *ptr;

    // Validao do ndice do SCU
    if ((scu=scuIndex(scu_id))>=N_SCU)  {
        buildStatusAnswer(MESSAGE_INVREQ);
        return;
    }

    // Condies especiais de funcionamento do sistema
    if (param==PARAM_SPECIAL)  {
        actuators[scu].sysStatus.pwm=usartBuffer.rx[DATA1] & 1;
        actuators[scu].sysStatus.calib=usartBuffer.rx[DATA1+1] & 1;
    }
    // Actuao sobre o sistema
    else  {
        // Preenchimento dos arrays internos com os novos dados recebidos (posio, velocidade ou corrente)
        for (srv=0; srv<N_SERVOS; ++srv)  {

            ptr=0;

            switch (actuators[scu].control[srv].type)  {
                // Controlo do Centro de Presso
                case COP_CONTROL:   switch (param)  {
                                        // Posio
                                        case PARAM_POSITION:    ptr=&refControl[scu].cop[srv].position;
                                                                break;
                                        // Velocidade
                                        case PARAM_VELOCITY:    ptr=&refControl[scu].cop[srv].velocity;
                                                                break;
                                    }
                                    break;
                // Controlo de inclinao
                case INC_CONTROL:   switch (param)  {
                                        // Posio
                                        case PARAM_POSITION:    ptr=&refControl[scu].inc[srv].position;
                                                                break;
                                        // Velocidade
                                        case PARAM_VELOCITY:    ptr=&refControl[scu].inc[srv].velocity;
                                                                break;
                                    }
                                    break;
                // Controlo de giroscpios
                case GIRO_CONTROL:  switch (param)  {
                                        // Posio
                                        case PARAM_POSITION:    ptr=&refControl[scu].giro[srv].position;
                                                                break;
                                        // Velocidade
                                        case PARAM_VELOCITY:    ptr=&refControl[scu].giro[srv].velocity;
                                                                break;
                                    }
                                    break;
                // Default: controlo dos servos
                default:            switch (param)  {
                                        // Posio
                                        case PARAM_POSITION:    ptr=&refControl[scu].servo[srv].position;
                                                                break;
                                        // Velocidade
                                        case PARAM_VELOCITY:    ptr=&refControl[scu].servo[srv].velocity;
                                                                break;
                                    }
            }

            if (ptr!=0)  *ptr=usartBuffer.rx[DATA1+srv];
        }
    }

    // Envio de uma mensagem indicadora que a operao pedida teve sucesso!
    buildStatusAnswer(MESSAGE_SUCESS);
}

//*****************************************************************************
// Aplicao de uma nova posio, velocidade ou corrente a todos os servos de uma junta
//
static void newJointControl(byte scu_id, byte param)
{
    byte scu, servo;

    // Validao do ndice do SCU
    if ((scu=scuIndex(scu_id))>=N_SCU)  {
        buildStatusAnswer(MESSAGE_INVREQ);
        return;
    }

    // Preenchimento dos arrays internos com os novos dados recebidos (posio, velocidade ou corrente)
    for (servo=0; servo<N_SERVOS; ++servo)
        switch (param)  {
            // Parmetro proporcional
            case PARAM_KI:          actuators[scu].control[servo].ki=usartBuffer.rx[DATA1+servo];
                                    break;
            // Parmetro derivativo
            case PARAM_KP:          actuators[scu].control[servo].kp=usartBuffer.rx[DATA1+servo];
                                    break;
            // Parmetro integrativo
            case PARAM_KD:          actuators[scu].control[servo].kd=usartBuffer.rx[DATA1+servo];
                                    break;
            // Parmetro on/off dos controladores
            case PARAM_CONTROLON:   actuators[scu].control[servo].type=usartBuffer.rx[DATA1+servo] & 0b11;
                                    break;
        }

    // Envio de uma mensagem indicadora que a operao pedida teve sucesso!
    buildStatusAnswer(MESSAGE_SUCESS);
}

//*****************************************************************************
// Envio do contedo do buffer externo pela USART
//
static void sendExtBuffer(byte typePacket)
{
    byte i;

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;
    usartBuffer.tx[TYPE]=typePacket;

    // Cpia dos dados do buffer externo para o interno
    for (i=0; i<USART_EXTBUFF_LENGTH; ++i)  {
        usartBuffer.tx[DATA1+i]=extBuffer.tx[i];    // Transferncia da informao
        extBuffer.tx[i]=0;                          // Clear do buffer origem
    }

    // A flag read indica que os dados foram lidos pela USART
    extBuffer.read=true;

    // Reset do estado de erro do CAN
    canClearStatus();
}

//*****************************************************************************
// Leitura sensorial de um dos SCU's
//
static void sendSensorsInfo(byte scu_id, byte param)
{
    // Validao do ndice do SCU
    if (scuIndex(scu_id)>=N_SCU)  {
        buildStatusAnswer(MESSAGE_INVREQ);
        return;
    }

    if (param==PARAM_SPECIAL)  sendSpecialInfo(scu_id);
    else                       sendJointInfo(scu_id,param);
}

//*****************************************************************************
// Leitura de um parmetro de todos os servos de uma junta
//
static void sendJointInfo(byte scu_id, byte param)
{
    byte i, servo, scu=scuIndex(scu_id);
    byte typePacket=typePacketSensors(OP_READ_SENSORS,scu_id,param);

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;                                // Start of Frame
    usartBuffer.tx[TYPE]=typePacket;                                    // Tipo da mensagem

    // Estado dos servos
    // Preenchimento do buffer de transmisso com as informaes sensoriais da junta
    for(servo=0; servo<N_SERVOS; ++servo)
        switch (param)  {
            // Posio
            case PARAM_POSITION:    usartBuffer.tx[DATA1+servo]=sensors[scu].servo[servo].position;
                                    break;
            // Velocidade
            case PARAM_VELOCITY:    usartBuffer.tx[DATA1+servo]=sensors[scu].servo[servo].velocity;
                                    break;
            // Corrente
            case PARAM_CURRENT:     usartBuffer.tx[DATA1+servo]=sensors[scu].servo[servo].current;
                                    break;
            // Outros
            default:                usartBuffer.tx[DATA1+servo]=0;
        }

    // Pacote de estado do sistema
    usartBuffer.tx[DATA1+N_SERVOS]=sensors[scu].sysStatus;

    // Preenchimento com zeros do resto da mensagem
    for (i=DATA1+N_SERVOS+1; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
}


//*****************************************************************************
// Leitura da matriz dos sensores de fora de um determinado p
//
static void sendSpecialInfo(byte scu_id)
{
    byte i, scu=scuIndex(scu_id);

    usartBuffer.tx[SOF ]=MESSAGE_SUCESS;                                    // Start Of Frame
    usartBuffer.tx[TYPE]=typePacketSensors(OP_READ_SENSORS,scu_id,PARAM_SPECIAL);  // Tipo da mensagem

    // Preenchimento do buffer de transmisso com os valores dos sensores de fora
    for(i=0; i<N_SPECIAL_SENSORS; ++i)  usartBuffer.tx[DATA1+i]=sensors[scu].special[i];

    // Preenchimento com zeros do resto da mensagem
    for (i=DATA1+N_SPECIAL_SENSORS; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
}


//*****************************************************************************
// Preparao de uma mensagem com a indicao de um status
//
static void buildStatusAnswer(byte stat)
{
    byte i;

    // O pacote SOF conter o status a enviar:
    usartBuffer.tx[SOF]=stat;

    // Preenchimento do resto da mensagem com o contedo da ltima recebida
    for (i=TYPE; i<USART_RECV_LENGTH; ++i)  usartBuffer.tx[i]=usartBuffer.rx[i];

    // Preenchimento com zeros do resto da mensagem
    for (i=USART_RECV_LENGTH; i<USART_SEND_LENGTH; ++i)  usartBuffer.tx[i]=0;
}

//////////////////////////////////////////////////////////
//   FUNES PARA FINALIZAO DA MENSAGEM DE RESPOSTA   //
//////////////////////////////////////////////////////////

//*****************************************************************************
// Adiciona o cdigo BCC ao fim de uma mensagem de transmisso
//
static void addBCC(void)
{
    byte i;
    byte bcc;

    for(bcc=0,i=0; i<USART_SEND_LENGTH; ++i)  bcc+=usartBuffer.tx[i];

    // A posio i corresponde  localizao do byte BCC
    usartBuffer.tx[USART_SEND_LENGTH]=bcc;
}

// Verifica se o cdigo BCC presente numa mensagem recebida est correcto
//
static bool checkBCC(void)
{
    byte i;
    byte bcc;

    for(bcc=0,i=0; i<USART_RECV_LENGTH; ++i)  bcc+=usartBuffer.rx[i];

    // A posio i corresponde  localizao do byte BCC
    return (usartBuffer.rx[USART_RECV_LENGTH]==bcc);
}
