#include "servo.h"
#include "pic2.h"
#include "can.h"
#include "data.h"

// Estrutura com os dados do controlador
typedef struct {
    double angle;               // ngulo a solicitar
    signed int  error_prev;     // Erro do ciclo anterior
    signed char y_prev;         // Sinal de feedback (potencimetro) do ciclo anterior
    signed char y_prev2;        // Sinal de feedback de h 2 ciclos atrs
} struct_controller;

// Estrutura com os dados do controlador
typedef struct  {
    bool errortol;                    // Deve tolerar pequenos erros?
    signed char actualPosition;      // Posio final a atingir
    signed char actualCoP;          // Centro de Presso referncia
    struct_controller controller;   // Dados do controlador
} struct_control;
static struct_control servoControl[N_SERVOS];      // Dados dos actuadores

// Estrutura utilizada para a calibrao dinmica (dynamicServoCalib function)
struct {
    signed char pos0;           // Posio original
    signed int deltamax;        // Mximo deslocamento
} dynCalib[N_SERVOS];

// Execuo da calibrao dos servos
//
void calibration(void)
{
    byte state, i, count;
    signed int med;
    unsigned int special_sum[N_SPECIAL_SENSORS];
    signed int position_sum[N_SERVOS];
    signed char posReq[N_SERVOS];

    // If going to calculate the multiplier factor, wait 3 seconds
    wait(1000);

    // ___ CALIBRAO DOS SENSORES DE FORA ____________________________
    // Inicializao das variveis locais
    for (i=0; i<N_SPECIAL_SENSORS; ++i)  {
        special0[i].offset=0;
        special0[i].slope=100;
    }

    // Offset and slope correction of the pressure sensors
    // In this situation, only the offset calibration is done (state<=1)
    for (state=1; state<=1; ++state)  {

        // Clearing the sum array
        for (i=0; i<N_SPECIAL_SENSORS; ++i)  special_sum[i]=0;

        // Execuo de 25 medidas
        for (count=0; count<25; ++count)  {
    		// Esperando pela passagem de um tick;
            waitTick();
    
            // Acumulao da sada dos sensores de fora
            for (i=0; i<N_SPECIAL_SENSORS; ++i)
                special_sum[i] += sensors.special[i];
        }
        // Correco dos desvios dos sensores de fora
        for (i=0; i<N_SPECIAL_SENSORS; ++i)  {
            // Medium read value
            med = (signed int) (special_sum[i]/25);

            switch (state)  {
            // Offset determinado
            case 1: special0[i].offset = (signed int) 128 - med;
                    break;
            // Calculing Slope
            // After the 3 seconds the foot must be located over the ground!
            case 2: special0[i].slope = (signed int) 128*100/(signed int) med;
                    break;
            }
        }
    }

    // ___ CALIBRAO DOS POTENCIMETROS DOS SERVOS ____________________
    // Inicializao de variveis relativas aos servos e activao do PWM
    for (i=0; i<N_SERVOS; ++i)  {

        // Inicializao da estrutura para calibrao dinmica (dynamicServoCalib function)
        dynCalib[i].pos0=posReq[i];
        dynCalib[i].deltamax=0;

        // Inicializao do array
        position_sum[i]=0;

        // Posio solicitada
        posReq[i]=servo_interface[i].actuation;

        // Activao do sinal de PWM
        statusPWM(true,i);

        // Wait some time
        wait(500);
    }

    // Execuo de 25 medidas
    for (count=0; count<25; ++count)  {
        // Esperando pela passagem de um tick;
        waitTick();

        // Acumulao das posies medidas
        for (i=0; i<N_SERVOS; ++i)
            position_sum[i] += servo_interface[i].position;
    }

    // Correco dos desvios da medio do potencimetro
    for (i=0; i<N_SERVOS; ++i)  {
        med = (signed int) (position_sum[i]/25);
        origin[i].offset += (signed int) posReq[i]-med;
    }
    // Calibrao efectuada!
}

// Calibrao do declive para clculo da posio angular
bool dynamicServoCalib(signed char posact, byte srv)
{
    signed int  delta;
    signed long offsetnew, slopenew;
    signed char pos0=dynCalib[srv].pos0;                    // Posio original (da ltima calibrao esttica)

    unsigned char str[]="\r\nNova Calibracao: Servo=";
    unsigned char strslope[]="     Slope=";
    unsigned char stroffset[]="  Offset=";

    // Verificando se est autorizada a calibrao dinmica
    if (!(sensors.sysStatus.calib=actuators.sysStatus.calib))  return false;

    // Apenas  feita a calibrao dinmica, se o controlo est em malha aberta.
    if (actuators.control[srv].type!=NO_CONTROL)  return false;

    // Diferena do deslocamento relativamente  posio original (mdulo)
    if ((delta=(signed int) pos0-(signed int) posact)<0)  delta*=-1;

    // Apenas faz calibrao dinmica se o deslocamento  superior ou igual a 10 e se
    // a diferena angular desde a posio original for mxima
    if (delta<10 || delta<=dynCalib[srv].deltamax)  return false;

    // Actualizao do mximo deslocamento
    dynCalib[srv].deltamax=delta;

    // Novo declive
    slopenew=(signed long) origin[srv].slope*((signed long) pos0-(signed long) posact)/((signed long) pos0-(signed long) servo_interface[srv].position);

    // Novo offset
    offsetnew=slopenew*((signed long) origin[srv].offset-(signed long) pos0)/(signed long) origin[srv].slope+(signed long) pos0;

    // Actualizao do declive e offset
    origin[srv].slope =(signed int) slopenew;
    origin[srv].offset=(signed int) offsetnew;
/*
    // Apresentao dos resultados finais
    sendStr(str);
    sendVal(srv,10);
    sendStr(strslope);
    sendValSign(origin[srv].slope,10);
    sendStr(stroffset);
    sendValSign(origin[srv].offset,10);
*/
    return true;
}


// Movimentao do servo para a posio solicitada
//
signed char servoRequest(signed char position, byte srv, byte control_type)
{
    // Correco da posio solicitada
    position=limitPosition(position);

    // Posio a solicitar
    if (control_type==NO_CONTROL || control_type==LOCOMOTION_CONTROL)
        servoControl[srv%N_SERVOS].actualPosition=limitPosition(position);  // Posio desejada
    else
        servoControl[srv%N_SERVOS].actualCoP=position;

    // Tolerncia a erros
    servoControl[srv%N_SERVOS].errortol=true;       // Tolaerancia a pequenos erros de posio

    return position;
}

// ==========================================================================================
// Funes temporais
// ==========================================================================================

// Inicializao do Controlador
//
void initController(void)
{
    byte srv;
    struct_controller *cont;

    for (srv=0; srv<N_SERVOS; ++srv)  {

        cont = &(servoControl[srv].controller);

        servoControl[srv].errortol=true;
        servoControl[srv].actualPosition=servo_interface[srv].actuation;
        cont->angle=(double) servoControl[srv].actualPosition;
        cont->error_prev=0;
        cont->y_prev=0;
        cont->y_prev2=0;
    }
}

static void openloopControl(byte servo);
static void locomotionControl(byte servo);
static void inclinControl(byte servo, signed int copinc[3]);
static void jacobianControl(byte servo, signed int copinc[3], signed int Jt[N_SERVOS][N_SERVOS]);

void controller(void)
{
    byte servo;
    signed int copinc[3];
    signed int Jt[N_SERVOS][N_SERVOS];

    // Clculo da matriz Jacobiana
    jacobianMatrix(Jt);

    for (servo=0; servo<N_SERVOS; ++servo)  {

        // Correco do tipo de controlo a usar
        if (actuators.control[servo].type==LOCOMOTION_CONTROL && actuators.control[servo].kp==0
            && actuators.control[servo].kd==0 && actuators.control[servo].ki==0)
            actuators.control[servo].type=NO_CONTROL;

        // Controlo em malha aberta
        openloopControl(servo);

        // Controlo de posio em malha fechada
        locomotionControl(servo);

        // Clculo do centro de presso (em dcimas de mm)
        calcInc(copinc);
        // Controlo das foras de reaco
        inclinControl(servo,copinc);

        // Clculo do centro de presso (em dcimas de mm)
        calcCoP(copinc);
        // Controlo de equilbrio
        jacobianControl(servo,copinc,Jt);
    }
}

// Limitao do controlo de posio
static signed char limitControlPosition(byte servo);

static void openloopControl(byte servo)
{
    // Se o controlo em malha aberta no estiver contemplado, abandonar!
    if (actuators.control[servo].type!=NO_CONTROL)  return;

    // CONTROLO EM MALHA ABERTA
    // Passagem da posio solicitada  a plicar
    servoControl[servo].controller.angle=(double) servoControl[servo].actualPosition;
    
    // Aplicao do movimento
    servoActuation(limitControlPosition(servo),servo);
}


#define FILTER_N  1.0           // Nvel de filtragem da componente derivativa

static void locomotionControl(byte servo)
{
    signed char posReference, y_now;
    signed int error, error_dif, y_vardif;
    double deltaP, deltaI, deltaD, kp, ki, kd;
    struct_controller *cont=&(servoControl[servo].controller);  // Ponteiro para a estrutura de controlo do servo

    posReference=servoControl[servo].actualPosition;       // Posio desejada (referncia)
    y_now = servo_interface[servo].position;              // Posio efectiva do servo
    
    // Erro actual
    error=(signed int) posReference - (signed int) y_now;

    // Tolerancia de pequenos erros
    if (servoControl[servo].errortol)  {
        // banda morta
        if (error>=-2 && error<=2)  error= 0;
        else if (error== 3)         error= 2;
        else if (error==-3)         error=-2;
    }
    
    // Diferena do erro para a aco proporcional e variao da diferena da sada para a aco derivativa
    error_dif= error-cont->error_prev;
    y_vardif =(signed int) y_now-2*(signed int) cont->y_prev+(signed int) cont->y_prev2;
    
    // Actualizao das sadas/erros anteriores
    cont->y_prev2=cont->y_prev;
    cont->y_prev =y_now;
    cont->error_prev=error;
    
    // Se o controlo de posio em malha fechada no for contemplado passar ao servo seguinte!
    if (actuators.control[servo].type!=LOCOMOTION_CONTROL)  return;

    // CONTROLO EM MALHA FECHADA DA LOCMOO

    ki = (double) actuators.control[servo].ki/(double) K_FACTOR;
    kp = (double) actuators.control[servo].kp/(double) K_FACTOR;
    kd = (double) actuators.control[servo].kd/(double) K_FACTOR;
    
    // Componentes do compensador
    deltaI = (double) error*ki;                     // Compensao integrativa
    deltaP = (double) error_dif*kp;                 // Compensao proporcional
    deltaD = (double) y_vardif*kd;                  // Compensao derivativa
    
    // Clculo da nova posio a aplicar
    cont->angle += deltaI + deltaP + deltaD;
    
    // Aplicao do movimento
    servoActuation(limitControlPosition(servo),servo);
}


// _____ CONTROLADOR DE EQUILBRIO PROPORCIONAL _______________________________________
// Processamento e aplicao da nova velocidade a aplicar
static void applyVelocity(byte servo, double velocity);

#define KP_FORCEY    0.1
#define KP_FORCEX    0.05

static void inclinControl(byte servo, signed int inc[3])
{
    double delta;
	double K = (double) actuators.control[servo].ki;

    // Se o tipo de controlo a aplicar no for de reaco, abandonar funo!
    if (actuators.control[servo].type!=INCLIN_CONTROL)  return;

    if (servo==0)  {
        // COMPENSAO DO CENTRO DE PRESSO NO EIXO XX
        // Clculo da variao de posio a aplicar
        delta = (double) servo_signal[servo]*inc[0]*2.5*K*1e-2;
    
        // Aplicao da nova velocidade ao servo 0
        applyVelocity(servo,delta);
    }
    else if (servo==1)  {
        // COMPENSAO DO CENTRO DE PRESSO NO EIXO YY
        // Actualizao da posio do servo
        delta = (double) servo_signal[servo]*inc[1]*2.5*K*1e-2;
    
        // Aplicao da nova velocidade ao servo 1
        applyVelocity(servo,delta);
    }
}

// _____ CONTROLADOR DE EQUILBRIO PELO JACOBIANO _____________________________________

static const signed char K_error[3]={10, 10, -20};      // Valor normalizado==10

static void jacobianControl(byte servo, signed int cop[3], signed int Jt[N_SERVOS][N_SERVOS])
{
    double delta;
    double K=(double) actuators.control[servo].ki;
    signed int copx_error, copy_error, copz_error;

    // Se o tipo de controlo no for de equlbrio, sair!
    if (actuators.control[servo].type!=BALANCE_CONTROL)  return;

    copx_error=((signed int) servoControl[0].actualCoP-cop[0])*K_error[0]/(signed int) 10;
    copy_error=((signed int) servoControl[1].actualCoP-cop[1])*K_error[1]/(signed int) 10;
    copz_error=((signed int) servoControl[2].actualCoP-cop[2])*K_error[2]/(signed int) 10;

    // Clculo do incremento de posio
    delta = (double) Jt[servo][0]*copx_error + (double) Jt[servo][1]*copy_error + (double) Jt[servo][2]*copz_error;
    // Ajuste de Delta
    delta = servo_signal[servo]*delta*K*1e-7;

    // Aplicao do incremento
    applyVelocity(servo,delta);
}


#define MAX_SPEED   50                              // Mxima velocidade (graus/s)
#define MAX_POSVAR  ((double) (MAX_SPEED)*(double) TICK_TIME/1000.0)

// Processamento e aplicao da nova velocidade a aplicar
static void applyVelocity(byte servo, double velocity)
{
    // Limitao de velocidade
    if (velocity<-MAX_POSVAR)      velocity = -MAX_POSVAR;
    else if (velocity>MAX_POSVAR)  velocity =  MAX_POSVAR;

    // Actualizao da posio do servo
    servoControl[servo].controller.angle += velocity;

    // Aplicao da nova posio
    servoActuation(limitControlPosition(servo),servo);
}


// Limitao da posio a aplicar  sada do controlador.
// O servo a limitar deve ser fornecido como parmetro
static signed char limitControlPosition(byte servo)
{
    struct_controller *cont;

    // Ponteiro para a estrutura de controlo correspondente
    cont = &(servoControl[servo].controller);

    // Limitao da posio a solicitar
    if ((signed int) cont->angle<-SERVO_RANGE)      cont->angle=-(double) SERVO_RANGE;
    else if ((signed int) cont->angle>SERVO_RANGE)  cont->angle= (double) SERVO_RANGE;

    return ((signed char) cont->angle);
}
